import unittest

from mqtt_homeassistant_utils import HASensor, HASensorEnergy, HASensorTemperature, HASensorBattery

class TestHASensor(unittest.TestCase):
    def test_mandatory_attributes(self):
        testDict = HASensor(name="MyTestSensor", node_id="justATest").to_dict()

        #check for attributes
        self.assertIn("name", testDict)
        self.assertEqual(testDict["name"], "MyTestSensor")

        self.assertNotIn("component", testDict)
        self.assertNotIn("node_id", testDict)
        self.assertNotIn("discovery_prefix", testDict)


        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "justATest_mytestsensor")

        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "justATest/values")


    def test_mandatory_missing(self):
        with self.assertRaises(TypeError):
            mySensor = HASensor(node_id="justATest")

    def test_own_uniqueid(self):
        testDict = HASensor(name="MyTestSensor", node_id="justATest", unique_id="uid").to_dict()
        
        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "uid")

    def test_own_statetopic(self):
        testDict = HASensor(name="MyTestSensor", node_id="justATest", state_topic="state/topic").to_dict()
        
        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "state/topic")

class TestHASensorEnergy(unittest.TestCase):
    def test_default_paramters(self):
        testDict = HASensorEnergy(name="MyTestSensor", node_id="justATest").to_dict()

        #check for attributes
        self.assertIn("name", testDict)
        self.assertEqual(testDict["name"], "MyTestSensor")

        self.assertNotIn("component", testDict)
        self.assertNotIn("node_id", testDict)
        self.assertNotIn("discovery_prefix", testDict)

        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "justATest_mytestsensor")

        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "justATest/values")

        self.assertIn("device_class", testDict)
        self.assertEqual(testDict["device_class"], "energy")

        self.assertIn("state_class", testDict)
        self.assertEqual(testDict["state_class"], "measurement")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "kWh")

    def test_override_paramters(self):
        testDict = HASensorEnergy(name="MyTestSensor", node_id="justATest", state_class="increasing", unit_of_measurement="mWh").to_dict()

        #check for attributes
        self.assertIn("state_class", testDict)
        self.assertEqual(testDict["state_class"], "increasing")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "mWh")

class TestHASensorTemperature(unittest.TestCase):
    def test_default_paramters(self):
        testDict = HASensorTemperature(name="MyTestSensor", node_id="justATest").to_dict()

        #check for attributes
        self.assertIn("name", testDict)
        self.assertEqual(testDict["name"], "MyTestSensor")

        self.assertNotIn("component", testDict)
        self.assertNotIn("node_id", testDict)
        self.assertNotIn("discovery_prefix", testDict)

        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "justATest_mytestsensor")

        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "justATest/values")

        self.assertIn("device_class", testDict)
        self.assertEqual(testDict["device_class"], "temperature")

        self.assertIn("state_class", testDict)
        self.assertEqual(testDict["state_class"], "measurement")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "°C")

    def test_override_paramters(self):
        testDict = HASensorTemperature(name="MyTestSensor", node_id="justATest", state_class="increasing", unit_of_measurement="°F").to_dict()

        #check for attributes
        self.assertIn("state_class", testDict)
        self.assertEqual(testDict["state_class"], "increasing")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "°F")

class TestHASensorEnergy(unittest.TestCase):
    def test_default_paramters(self):
        testDict = HASensorBattery(name="MyTestSensor", node_id="justATest").to_dict()

        #check for attributes
        self.assertIn("name", testDict)
        self.assertEqual(testDict["name"], "MyTestSensor")

        self.assertNotIn("component", testDict)
        self.assertNotIn("node_id", testDict)
        self.assertNotIn("discovery_prefix", testDict)


        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "justATest_mytestsensor")

        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "justATest/values")

        self.assertIn("device_class", testDict)
        self.assertEqual(testDict["device_class"], "battery")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "%")

    def test_override_paramters(self):
        testDict = HASensorBattery(name="MyTestSensor", node_id="justATest", state_class="increasing", unit_of_measurement="mAh").to_dict()

        #check for attributes
        self.assertIn("state_class", testDict)
        self.assertEqual(testDict["state_class"], "increasing")

        self.assertIn("unit_of_measurement", testDict)
        self.assertEqual(testDict["unit_of_measurement"], "mAh")        

if __name__ == '__main__':
    unittest.main()    


from dataclasses import dataclass
from typing import Optional, List, ClassVar

from .base import __HAEntry

@dataclass(kw_only=True)
class HAClimate(__HAEntry):
    component: ClassVar[str] = "climate"

    action_template: Optional[str] = None
    action_topic: Optional[str] = None
    current_humidity_template: Optional[str] = None
    current_humidity_topic: Optional[str] = None
    current_temperature_template: Optional[str] = None
    current_temperature_topic: Optional[str] = None
    fan_mode_command_template: Optional[str] = None
    fan_mode_command_topic: Optional[str] = None
    fan_mode_state_template: Optional[str] = None
    fan_mode_state_topic: Optional[str] = None
    fan_modes: Optional[List[str]]
    initial: Optional[float] = None
    max_humidity: Optional[int] = None
    max_temp: Optional[float] = None
    min_humidity: Optional[int] = None
    min_temp: Optional[float] = None
    mode_command_template: Optional[str] = None
    mode_command_topic: Optional[str] = None
    mode_state_template: Optional[str] = None
    mode_state_topic: Optional[str] = None
    modes: Optional[List[str]]
    optimistic: Optional[bool] = None
    payload_off: Optional[str] = None
    payload_on: Optional[str] = None
    power_command_template: Optional[str] = None
    power_command_topic: Optional[str] = None
    precision: Optional[float] = None
    preset_mode_command_template: Optional[str] = None
    preset_mode_command_topic: Optional[str] = None
    preset_mode_state_topic: Optional[str] = None
    preset_mode_value_template: Optional[str] = None
    preset_modes: Optional[List[str]]
    retain: Optional[bool] = None
    swing_mode_command_template: Optional[str] = None
    swing_mode_command_topic: Optional[str] = None
    swing_mode_state_template: Optional[str] = None
    swing_mode_state_topic: Optional[str] = None
    swing_modes: Optional[List[str]]
    target_humidity_command_template: Optional[str] = None
    target_humidity_command_topic: Optional[str] = None
    target_humidity_state_template: Optional[str] = None
    target_humidity_state_topic: Optional[str] = None
    temp_step: Optional[float] = None
    temperature_command_template: Optional[str] = None
    temperature_command_topic: Optional[str] = None
    temperature_high_command_template: Optional[str] = None
    temperature_high_command_topic: Optional[str] = None
    temperature_high_state_template: Optional[str] = None
    temperature_high_state_topic: Optional[str] = None
    temperature_low_command_template: Optional[str] = None
    temperature_low_command_topic: Optional[str] = None
    temperature_low_state_template: Optional[str] = None
    temperature_low_state_topic: Optional[str] = None
    temperature_state_template: Optional[str] = None
    temperature_state_topic: Optional[str] = None
    temperature_unit: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()


from dataclasses import dataclass, asdict, field
from typing import Optional, Union, Any, Dict, Tuple, ClassVar, Type
from enum import Enum

import json
import paho.mqtt.client as mqtt
import re

def find_class_var(klass: Type, var_name: str) -> str:
    """Recursively search for a class variable in the given class and its parent classes."""
    if hasattr(klass, var_name):
        return getattr(klass, var_name)
    else:
        for base_class in klass.__bases__:
            result = find_class_var(base_class, var_name)
            if result is not None:
                return result
    return None


@dataclass
class __BaseDataClass():
    """ Base function for all Entries, like to_json """
    def to_json(self) -> json:
        data_dict = self.to_dict()
        json_data = json.dumps(data_dict, indent=2)
        return json_data

    def remove_null_values(self, obj: Union[Dict[str, Any], Any]) -> Union[Dict[str, Any], Any]:
        if isinstance(obj, list):
            return [self.remove_null_values(item) for item in obj]
        elif isinstance(obj, dict):
            return {key: self.remove_null_values(value) for key, value in obj.items() if value is not None}
        elif isinstance(obj, Enum): 
            return obj.value if obj.value is not None else None
        elif hasattr(obj, '__dict__'):
            # Dataclass oder Objekt mit __dict__ (z.B. NamedTuple)
            return self.remove_null_values(asdict(obj))
        else:
            return obj
        
    def to_dict(self) -> dict:
        data_dict = asdict(self)

        #remove internal attributes
        try:
            del data_dict["node_id"], data_dict["discovery_prefix"], data_dict["component"]
        except KeyError:
            pass
        
        data_dict = self.remove_null_values(data_dict)

        return data_dict
        
@dataclass(kw_only=True)
class HADevice(__BaseDataClass):
    """ Device Subentry for all integrations """
    configuration_url: Optional[str] = None
    connections: Optional[Tuple[Tuple[str, str]]] = None
    hw_version: Optional[str] = None
    identifiers: Union[str, Tuple[str], None] = None
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    name: Optional[str] = None
    suggested_area: Optional[str] = None
    sw_version: Optional[str] = None
    via_device: Optional[str] = None

@dataclass(kw_only=True)
class HAAvailability(__BaseDataClass):
    topic: str

    payload_available: Optional[str] = None
    payload_not_available: Optional[str] = None
    value_template: Optional[str] = None


@dataclass(kw_only=True)
class __HAEntry(__BaseDataClass):
    name: str
    node_id: str

    component: ClassVar[str] = ""
    discovery_prefix: str = "homeassistant"
    
    availability: Optional[Tuple[HAAvailability]] = None
    availability_mode: Optional[str] = None
    device: Optional[HADevice] = None
    enabled_by_default: Optional[bool] = None
    encoding: Optional[str] = None
    entity_category: Optional[str] = None
    icon: Optional[str] = None
    json_attributes_template: Optional[str] = None
    json_attributes_topic: Optional[str] = None
    
    object_id: Optional[str] = None
    qos: Optional[int] = None
    unique_id: Optional[str] = None
    value_template: Optional[str] = None

    def publish(self, mqttClient: mqtt.Client):
        
        component = find_class_var(self.__class__, "component")
        if component is None or component == "":
            raise Exception("No component given")
        
        topic = (self.discovery_prefix + "/" + 
                 self.component + "/" + 
                 self.node_id + "/" + 
                 self.unique_id + "/config"
                )
        qos = (self.qos if self.qos else 0)
        
        mqttClient.publish(topic, self.to_json(), qos=qos, retain=True)

    def cleanName(self):
        return re.sub(r'[^A-Za-z]', '', self.name).lower()
        
    def __post_init__(self):
        if self.unique_id is None:
            self.unique_id = self.node_id + "_" + self.cleanName()
        if self.value_template is None:
            self.value_template = "{{ value_json." + self.cleanName() + " }}"
                



from dataclasses import dataclass
from typing import Optional, List, ClassVar

from .base import __HAEntry
  
@dataclass(kw_only=True)
class HASelect(__HAEntry):
    component: ClassVar[str] = "select"

    command_topic: str
    options: List[str]
    
    command_template: Optional[str] = None
    optimistic: Optional[bool] = None
    retain: Optional[bool] = None
    state_topic: Optional[str] = None
    unique_id: Optional[str] = None
    
    def __post_init__(self):
        super().__post_init__()

        self.component = "select"

        if self.state_topic is None:
            self.state_topic = self.node_id + "/values"
    


import unittest

from mqtt_homeassistant_utils import HABinarySensor

class TestHABinarySensor(unittest.TestCase):
    def test_mandatory_attributes(self):
        testDict = HABinarySensor(name="MyTestSensor", node_id="justATest").to_dict()

        #check for attributes
        self.assertIn("name", testDict)
        self.assertEqual(testDict["name"], "MyTestSensor")

        self.assertNotIn("component", testDict)
        self.assertNotIn("node_id", testDict)
        self.assertNotIn("discovery_prefix", testDict)

        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "justATest_mytestsensor")

        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "justATest/values")

    def test_mandatory_missing(self):
        
        with self.assertRaises(TypeError):
            myBinarySensor = HABinarySensor(node_id="justATest")

    def test_own_uniqueid(self):
        testDict = HABinarySensor(name="MyTestSensor", node_id="justATest", unique_id="uid").to_dict()
        
        self.assertIn("unique_id", testDict)
        self.assertEqual(testDict["unique_id"], "uid")

    def test_own_statetopic(self):
        testDict = HABinarySensor(name="MyTestSensor", node_id="justATest", state_topic="state/topic").to_dict()
        
        self.assertIn("state_topic", testDict)
        self.assertEqual(testDict["state_topic"], "state/topic")
        
if __name__ == '__main__':
    unittest.main()    


from dataclasses import dataclass
from typing import Optional, ClassVar
from enum import Enum

from .base import __HAEntry

class HADeviceClassBinarySensor(Enum):
    NONE = None
    BATTERY = "battery"
    BATTERY_CHARGING = "battery_charging"
    CARBON_MONOXIDE = "carbon_monoxide"
    COLD = "cold"
    CONNECTIVITY = "connectivity"
    DOOR = "door"
    GARAGE_DOOR = "garage_door"
    GAS = "gas"
    HEAT = "heat"
    LIGHT = "light"
    LOCK = "lock"
    MOISTURE = "moisture"
    MOTION = "motion"
    MOVING = "moving"
    OCCUPANCY = "occupancy"
    OPENING = "opening"
    PLUG = "plug"
    POWER = "power"
    PRESENCE = "presence"
    PROBLEM = "problem"
    RUNNING = "running"
    SAFETY = "safety"
    SMOKE = "smoke"
    SOUND = "sound"
    TAMPER = "tamper"
    UPDATE = "update"
    VIBRATION = "vibration"
    WINDOW = "window"
  
@dataclass(kw_only=True)
class HABinarySensor(__HAEntry):
    component: ClassVar[str] = "binary_sensor"

    device_class: Optional[HADeviceClassBinarySensor] = HADeviceClassBinarySensor.NONE
    expire_after: Optional[int] = None
    force_update: Optional[bool] = None
    off_delay: Optional[int] = None
    payload_off: Optional[str] = None
    payload_on: Optional[str] = None
    state_topic: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()

        if self.state_topic is None:
            self.state_topic = self.node_id + "/values"

 



from dataclasses import dataclass, field
from typing import Optional, ClassVar
from enum import Enum

from .base import __HAEntry

class HADeviceClassSensor(Enum):
    NONE = None
    APPARENT_POWER = "apparent_power"
    AQI = "aqi"
    ATMOSPHERIC_PRESSURE = "atmospheric_pressure"
    BATTERY = "battery"
    CARBON_DIOXIDE = "carbon_dioxide"
    CARBON_MONOXIDE = "carbon_monoxide"
    CURRENT = "current"
    DATA_RATE = "data_rate"
    DATA_SIZE = "data_size"
    DATE = "date"
    DISTANCE = "distance"
    DURATION = "duration"
    ENERGY = "energy"
    ENERGY_STORAGE = "energy_storage"
    ENUM = "enum"
    FREQUENCY = "frequency"
    GAS = "gas"
    HUMIDITY = "humidity"
    ILLUMINANCE = "illuminance"
    IRRADIANCE = "irradiance"
    MOISTURE = "moisture"
    MONETARY = "monetary"
    NITROGEN_DIOXIDE = "nitrogen_dioxide"
    NITROGEN_MONOXIDE = "nitrogen_monoxide"
    NITROUS_OXIDE = "nitrous_oxide"
    OZONE = "ozone"
    PH = "ph"
    PM1 = "pm1"
    PM25 = "pm25"
    PM10 = "pm10"
    POWER_FACTOR = "power_factor"
    POWER = "power"
    PRECIPITATION = "precipitation"
    PRECIPITATION_INTENSITY = "precipitation_intensity"
    PRESSURE = "pressure"
    REACTIVE_POWER = "reactive_power"
    SIGNAL_STRENGTH = "signal_strength"
    SOUND_PRESSURE = "sound_pressure"
    SPEED = "speed"
    SULPHUR_DIOXIDE = "sulphur_dioxide"
    TEMPERATURE = "temperature"
    TIMESTAMP = "timestamp"
    VOLATILE_ORGANIC_COMPOUNDS = "volatile_organic_compounds"
    VOLATILE_ORGANIC_COMPOUNDS_PARTS = "volatile_organic_compounds_parts"
    VOLTAGE = "voltage"
    VOLUME = "volume"
    VOLUME_STORAGE = "volume_storage"
    WATER = "water"
    WEIGHT = "weight"
    WIND_SPEED = "wind_speed"

@dataclass(kw_only=True)
class HASensor(__HAEntry):
    component: ClassVar[str] = "sensor"

    device_class: Optional[HADeviceClassSensor] = HADeviceClassSensor.NONE
    expire_after: Optional[int] = None
    force_update: Optional[bool] = None
    last_reset_value_template: Optional[str] = None
    suggested_display_precision: Optional[int] = None
    state_class: Optional[str] = None
    state_topic: Optional[str] = None
    unit_of_measurement: Optional[str] = None

    def __post_init__(self):
        super().__post_init__()        

        if self.state_topic is None:
            self.state_topic = self.node_id + "/values"

@dataclass(kw_only=True)
class HASensorEnergy(HASensor):
    device_class: HADeviceClassSensor = HADeviceClassSensor.ENERGY
    state_class: str = "measurement"
    unit_of_measurement: str = "kWh"
    
    def __post_init__(self):
        super().__post_init__()

@dataclass(kw_only=True)
class HASensorTemperature(HASensor):
    device_class: HADeviceClassSensor = HADeviceClassSensor.TEMPERATURE
    state_class: str = "measurement"
    unit_of_measurement: str = "°C"

    def __post_init__(self):
        super().__post_init__()

@dataclass(kw_only=True)
class HASensorBattery(HASensor):
    device_class: HADeviceClassSensor = HADeviceClassSensor.BATTERY
    unit_of_measurement: str = "%"

    def __post_init__(self):
        super().__post_init__()


import unittest

from mqtt_homeassistant_utils import HADevice, HAAvailability

class TestHADevice(unittest.TestCase):
    def test_hadevice(self):
        modelstring = "aNiceModel"
        testDict = HADevice(model=modelstring).to_dict()

        #check for attributes
        self.assertIn("model", testDict)
        self.assertEqual(testDict["model"], modelstring)

        self.assertNotIn("sw_version", testDict)

class TestHAAvailability(unittest.TestCase):
    def test_haavailability_mandatory(self):
        path_string = "path/to/topic"

        testDict = HAAvailability(topic=path_string).to_dict()

        #check for attributes
        self.assertIn("topic", testDict)
        self.assertEqual(testDict["topic"], path_string)

        self.assertNotIn("payload_available", testDict)

    def test_haavailability_mandatory_missing(self):
        
        with self.assertRaises(TypeError):
            myAv = HAAvailability(payload_available="off")

if __name__ == '__main__':
    unittest.main()    